# -*- coding: utf-8; mode: Python;
__author__ = 'Erdene-Ochir Tuguldur'

import os
import re
import sys
import queue
import threading
import time
from os.path import dirname, join
from collections import OrderedDict

#
# NVDA imports
#
import config
import nvwave
import gettext
import addonHandler
import languageHandler
import synthDriverHandler
from logHandler import log
from autoSettingsUtils.utils import StringParameterInfo
from autoSettingsUtils.driverSetting import DriverSetting, NumericDriverSetting, BooleanDriverSetting
from synthDriverHandler import SynthDriver, LanguageInfo, VoiceInfo, synthIndexReached, synthDoneSpeaking, changeVoice
# https://github.com/nvaccess/nvda/blob/master/source/speech/commands.py
from speech.commands import IndexCommand, CharacterModeCommand, LangChangeCommand, PitchCommand

LOCALE_DIR = os.path.join(os.path.dirname(__file__), '..', '..', 'locale')
TRANS = gettext.translation('nvda', localedir=LOCALE_DIR, languages=[languageHandler.getLanguage()], fallback=True)
_ = TRANS.gettext

#
# TTS imports
#
DEPENDENCIES_PATH = join(dirname(__file__), 'python_dependencies')
if DEPENDENCIES_PATH not in sys.path:
    sys.path.append(DEPENDENCIES_PATH)
os.environ['PHONEMIZER_ESPEAK_LIBRARY'] = join(dirname(__file__), 'espeak', 'libespeak-ng.dll')

from .tts_normalizer import init_normalizers
from .chimege_utils import ChimegeSynthesizer
from .nvda_synthesizer import NVDASynthesizer

tts_normalizer_config = {'noun_conj': 'noun_conj.json',
                         'symbols_file': 'symbols.json',
                         'abbreviation_file': 'mongolian_abbreviations.csv',
                         'emoji_list': 'emoji_list.json',
                         'soft_threshold': 150,
                         'hard_threshold': 300}
tts_normalizer_options = {'split_sentences': True,
                          'abbreviation_level': 'abbreviation',
                          'read_symbols': 'no_temdegt',
                          'read_emojis': True,
                          'use_phonemizer': True,
                          'dont_read_number_n': False,
                          'read_legal_number': False,
                          'read_roman_number': False,
                          'number_chunker': 'default'}
init_normalizers(tts_normalizer_config)


def _log_to_file(msg, file_name='Z:\\log.txt'):
    with open(file_name, 'at', encoding="utf-8") as f:
        f.write(str(msg) + '\n')


class TTSThread(threading.Thread):
    def __init__(self, tts_queue):
        super().__init__()
        self._tts_queue = tts_queue
        self.setDaemon(True)

    def run(self):
        while True:
            try:
                task = self._tts_queue.get()
                if task is None:
                    break
                else:
                    task()
            except:
                log.error("ChimegeSonur: error while executing a tts task", exc_info=True)
            self._tts_queue.task_done()
            time.sleep(0.005)


class TTSTask(object):

    def __init__(self, synthesizer, nvda_synthesizer, speech_sequence, on_index_reached, options={}):
        self._synthesizer = synthesizer
        self._nvda_synthesizer = nvda_synthesizer
        self._speech_sequence = speech_sequence
        self._on_index_reached = on_index_reached
        self._options = options

    def __call__(self):
        if self._synthesizer.cancel_flag.is_set():
            return

        chunk_length = self._options.get('vocoder_segment_size', 100)
        overlap_length = self._options.get('vocoder_segment_overlap', 5)
        read_multi_lines = self._options.get('read_multi_lines', False)
        number_chunker = self._options.get('number_chunker', 'default')
        number_n = self._options.get('number_n', True)
        roman_number = self._options.get('roman_number', True)
        read_abbreviation = self._options.get('read_abbreviation', True)

        _tts_normalizer_options = tts_normalizer_options.copy()
        _tts_normalizer_options['number_chunker'] = number_chunker
        _tts_normalizer_options['dont_read_number_n'] = not number_n
        _tts_normalizer_options['read_roman_number'] = roman_number
        _tts_normalizer_options['abbreviation_level'] = 'abbreviation' if read_abbreviation else 'letter'
        # _log_to_file(number_chunker)
        # _log_to_file(_tts_normalizer_options)

        if overlap_length not in [5, 10, 15]:
            overlap_length = 5
        char_mode = False
        # _log_to_file(self._speech_sequence)
        if read_multi_lines:
            is_say_all = True
            for command in self._speech_sequence:
                if isinstance(command, str) or isinstance(command, IndexCommand) or isinstance(command,
                                                                                               LangChangeCommand):
                    pass
                else:
                    is_say_all = False
                    break
            if is_say_all:
                new_speech_sequence, index_commands = [], []
                text = None
                if len(self._speech_sequence) > 0 and isinstance(self._speech_sequence[0], IndexCommand):
                    new_speech_sequence.append(self._speech_sequence[0])
                    self._speech_sequence = self._speech_sequence[1:]

                for command in self._speech_sequence:
                    if isinstance(command, str):
                        text = command if text is None else text.strip() + " " + command.strip()
                    elif isinstance(command, LangChangeCommand):
                        # if text is not None:
                        #    new_speech_sequence.append(text)
                        # text = None
                        # new_speech_sequence.append(command)
                        pass
                    elif isinstance(command, IndexCommand):
                        index_commands.append(command)
                if text is not None:
                    new_speech_sequence.append(text)
                new_speech_sequence = new_speech_sequence + index_commands
                self._speech_sequence = new_speech_sequence
        # _log_to_file("new")
        # _log_to_file(self._speech_sequence)
        for command in self._speech_sequence:
            if isinstance(command, str):
                text = command.strip()
                if not text or self._synthesizer.cancel_flag.is_set():
                    continue

                self._synthesizer.synthesize_multilingual(self._nvda_synthesizer,
                                                          text, _tts_normalizer_options,
                                                          chunk_length=chunk_length, overlap_length=overlap_length,
                                                          log_func=None)  # log_func=lambda msg: log.info(msg))
            elif isinstance(command, IndexCommand):
                index = command.index
                self._on_index_reached(index)
            elif isinstance(command, PitchCommand):
                offset = command.offset
                self._synthesizer.pitch_offset = offset / 100.0
            elif isinstance(command, CharacterModeCommand):
                char_mode = command.state
            elif isinstance(command, LangChangeCommand):
                pass
            else:
                log.warn("unsupported command: " + str(command))


class DoneSpeaking(object):

    def __init__(self, player, onIndexReached):
        self._player = player
        self._onIndexReached = onIndexReached

    def __call__(self):
        self._player.idle()
        self._onIndexReached(None)


class SynthDriver(SynthDriver):
    name = "ChimegeSonur"
    description = "ChimegeSonur"

    supportedSettings = [
        SynthDriver.VoiceSetting(),
        DriverSetting('enSynthesizerName', _("English synt&hesizer"),
                      defaultVal='none', availableInSettingsRing=False),
        SynthDriver.RateSetting(),
        SynthDriver.RateBoostSetting(),
        SynthDriver.PitchSetting(),
        SynthDriver.VolumeSetting(),
        NumericDriverSetting("vocoderSegmentSize", _("Vocoder segment siz&e"),
                             defaultVal=100, minVal=5, maxVal=150, minStep=5,
                             availableInSettingsRing=True),
        DriverSetting('vocoderSegmentOverlap', _("Vocoder se&gment overlap"),
                      defaultVal='5', availableInSettingsRing=False),
        DriverSetting('numberChunker', _("Number read &format"),
                      defaultVal='default', availableInSettingsRing=True),
        BooleanDriverSetting('numberN', _("Read number with N suffix"), defaultVal=True, availableInSettingsRing=True),
        BooleanDriverSetting('romanNumber', _("Read roman number"), defaultVal=True, availableInSettingsRing=True),
        BooleanDriverSetting('readMultiLines', _("Read multi line&s"), defaultVal=True, availableInSettingsRing=False),
        BooleanDriverSetting('readAbbreviation', _("Rea&d abbreviation"), defaultVal=True, availableInSettingsRing=False)
    ]

    supportedCommands = {
        IndexCommand,
        CharacterModeCommand,
        LangChangeCommand,
        PitchCommand
    }
    supportedNotifications = {synthIndexReached, synthDoneSpeaking}

    @classmethod
    def check(cls):
        for addon in addonHandler.getRunningAddons():
            if re.match("ChimegeSonur-voice.*", addon.name):
                return True
        return False

    def __init__(self):
        self._voice = "female1"
        self._rate = 50
        self._pitch = 50
        self._volume = 50
        self._rate_boost = False
        self._vocoder_segment_size = 100
        self._vocoder_segment_overlap = '5'
        self._number_chunker = 'default'
        self._number_n = True
        self._roman_number = True
        self._read_multi_lines = False
        self._read_abbreviation = True

        # find installed voices
        self._voices = {}
        for addon in addonHandler.getRunningAddons():
            if re.match("ChimegeSonur-voice.*", addon.name):
                voice_id = addon.name.split('-')[-1]
                self._voices[voice_id] = addon.path
        if len(self._voices) > 0:
            voice_ids = sorted(self._voices.keys())
            self._voice = voice_ids[0]

        self._player = nvwave.WavePlayer(channels=1, samplesPerSec=22050, bitsPerSample=16,
                                         outputDevice=config.conf["speech"]["outputDevice"], buffered=False)
        self._cancel_flag = threading.Event()

        self._nvda_synthesizer = NVDASynthesizer(self._player, self._cancel_flag)
        self._en_synthesizer_name = 'none'

        self._synthesizer = ChimegeSynthesizer(self._player, self._cancel_flag)
        self._tts_queue = queue.Queue()
        self._tts_thread = TTSThread(self._tts_queue)
        self._tts_thread.start()

    def terminate(self):
        self.cancel()
        self._tts_queue.put(None)
        self._tts_thread.join()
        self._player.close()
        if self._nvda_synthesizer is not None:
            self._nvda_synthesizer.terminate()

    def cancel(self):
        self._cancel_flag.set()
        task_list = []
        try:
            while True:
                task = self._tts_queue.get_nowait()
                self._tts_queue.task_done()
                if isinstance(task, DoneSpeaking):
                    task_list.append(task)
        except queue.Empty:
            pass
        for task in task_list:
            self._tts_queue.put(task)

        self._tts_queue.put(self._cancel_flag.clear)
        self._player.stop()
        self._tts_queue.join()

    def speak(self, speech_sequence):
        task = TTSTask(self._synthesizer, self._nvda_synthesizer, speech_sequence, self._on_index_reached,
                       {
                           'vocoder_segment_size': self._vocoder_segment_size,
                           'vocoder_segment_overlap': int(self._vocoder_segment_overlap),
                           'read_multi_lines': self._read_multi_lines,
                           'number_chunker': self._number_chunker,
                           'number_n': self._number_n,
                           'roman_number': self._roman_number,
                           'read_abbreviation': self._read_abbreviation
                       })
        self._tts_queue.put(task)
        self._tts_queue.put(DoneSpeaking(self._player, self._on_index_reached))

    def _on_index_reached(self, index):
        if index is not None:
            synthIndexReached.notify(synth=self, index=index)
        else:
            synthDoneSpeaking.notify(synth=self)

    def pause(self, switch):
        self._player.pause(switch)

    def _get_availableVoices(self):
        voice_ids = sorted(self._voices.keys())
        return OrderedDict({
            voice_id: VoiceInfo(voice_id, voice_id, 'mn_MN') for voice_id in voice_ids
        })

    def _get_rate(self):
        return self._rate

    def _set_rate(self, rate):
        self._rate = max(0, min(100, rate))
        chimege_speed = 1.0
        if self._rate < 50:
            chimege_speed = 0.5 + 0.01 * self._rate
        elif self._rate > 50:
            chimege_speed = 1.0 + 0.04 * (self._rate - 50)
        if self._rate_boost:
            chimege_speed = 2.0 * chimege_speed
        self._synthesizer.speed = chimege_speed

    def _get_pitch(self):
        return self._pitch

    def _set_pitch(self, pitch):
        self._pitch = max(0, min(100, pitch))
        chimege_pitch = 1.0
        if self._pitch < 50:
            chimege_pitch = 0.5 + 0.01 * self._pitch
        elif self._pitch > 50:
            chimege_pitch = 1.0 + 0.02 * (self.pitch - 50)
        self._synthesizer.pitch = chimege_pitch

    def _get_volume(self):
        return self._volume

    def _set_volume(self, volume):
        self._volume = max(0, min(100, volume))
        chimege_volume = 0.02 * self._volume
        self._synthesizer.volume = chimege_volume

    def _get_rateBoost(self):
        return self._rate_boost

    def _set_rateBoost(self, flag):
        self._rate_boost = flag
        self._set_rate(self._rate)

    def _get_voice(self):
        return self._voice

    def _set_voice(self, voice):
        self._synthesizer.load_voice_model(voice, self._voices[voice])
        self._synthesizer.current_voice = voice
        self._voice = voice

    def _get_vocoderSegmentSize(self):
        return self._vocoder_segment_size

    def _set_vocoderSegmentSize(self, size):
        self._vocoder_segment_size = max(5, min(150, size))

    def _get_vocoderSegmentOverlap(self):
        return self._vocoder_segment_overlap

    def _set_vocoderSegmentOverlap(self, overlap):
        self._vocoder_segment_overlap = overlap

    def _get_availableVocodersegmentoverlaps(self):
        overlaps = OrderedDict()
        overlaps["5"] = StringParameterInfo("5", "5")
        overlaps["10"] = StringParameterInfo("10", "10")
        overlaps["15"] = StringParameterInfo("15", "15")
        return overlaps

    def _get_numberChunker(self):
        return self._number_chunker

    def _set_numberChunker(self, number_chunker):
        self._number_chunker = number_chunker

    def _get_availableNumberchunkers(self):
        number_chunkers = OrderedDict()
        number_chunkers["default"] = StringParameterInfo("default", _("default"))
        number_chunkers["single"] = StringParameterInfo("single", _("digit"))
        number_chunkers["double"] = StringParameterInfo("double", _("two digits"))
        return number_chunkers

    def _get_readMultiLines(self):
        return self._read_multi_lines

    def _set_readMultiLines(self, flag):
        self._read_multi_lines = flag

    def _get_numberN(self):
        return self._number_n

    def _set_numberN(self, flag):
        self._number_n = flag

    def _get_readAbbreviation(self):
        return self._read_abbreviation

    def _set_readAbbreviation(self, flag):
        self._read_abbreviation = flag

    def _get_romanNumber(self):
        return self._roman_number

    def _set_romanNumber(self, flag):
        self._roman_number = flag

    def _get_enSynthesizerName(self):
        return self._en_synthesizer_name

    def _set_enSynthesizerName(self, synthesizer_name):
        self._en_synthesizer_name = synthesizer_name
        self._nvda_synthesizer.set_synthesizer(self._en_synthesizer_name)
        changeVoice(self, self._voice)

    def _get_availableEnsynthesizernames(self):
        synthesizers = OrderedDict()
        synthesizers['none'] = StringParameterInfo('none', 'Chimege')
        for synth_id, synth_display_name in synthDriverHandler.getSynthList():
            if synth_id in ['espeak', 'oneCore', 'vocalizer_expressive2', 'ibmeci']:
                synthesizers[synth_id] = StringParameterInfo(synth_id, synth_display_name)
        return synthesizers
