__author__ = 'Erdene-Ochir Tuguldur'

import os
import re
import wave
import time
import sonic
import contextlib
import numpy as np
import onnxruntime as ort

from .tts_normalizer import normalize_tts_text
from .lang_segmentation import segment_text_by_language
from .wav_synthesizer import synthesize_if_possible

_pad = '_'
_punctuation = '!\'(),.:;? '
_special = '-'
# _letters = 'АБВГДЕЁЖЗИЙКЛМНОӨПРСТУҮФХЦЧШЪЫЬЭЮЯабвгдеёжзийклмноөпрстуүфхцчшъыьэюя'
_letters = 'абвгдеёжзийклмноөпрстуүфхцчшъыьэюя'
symbols = [_pad] + list(_special) + list(_punctuation) + list(_letters)

# Mappings from symbol to numeric ID and vice versa:
_symbol_to_id = {s: i for i, s in enumerate(symbols)}
_id_to_symbol = {i: s for i, s in enumerate(symbols)}

# Regular expression matching whitespace:
_whitespace_re = re.compile(r'\s+')


def lowercase(text):
    return text.lower()


def collapse_whitespace(text):
    return re.sub(_whitespace_re, ' ', text)


def _should_keep_symbol(s):
    # TODO: do i really need this?
    return s in _symbol_to_id and s != '_'


def text_to_sequence(text):
    text = lowercase(text)
    text = collapse_whitespace(text)
    return [_symbol_to_id[s] for s in text if _should_keep_symbol(s)]


def get_wav(file_name):
    with contextlib.closing(wave.open(os.path.join(os.path.dirname(__file__), 'wavs', file_name), 'rb')) as wf:
        num_channels = wf.getnchannels()
        assert num_channels == 1
        sample_width = wf.getsampwidth()
        assert sample_width == 2
        sample_rate = wf.getframerate()
        assert sample_rate == 22050
        pcm_data = wf.readframes(wf.getnframes())
        return pcm_data


def run_onnx(ort_session, input_vals):
    ort_inputs = {name.name: val for name, val in zip(ort_session.get_inputs(), input_vals)}
    ort_outs = ort_session.run(None, ort_inputs)
    return ort_outs[0]


class ChimegeSynthesizer(object):

    def __init__(self, player, cancel_flag, sample_rate=22050):
        self.voices = {}
        self.current_voice = None
        self.pitch = 1.0
        self.pitch_offset = 0.0
        self.speed = 1.0
        self.volume = 1.0
        self.player = player
        self.cancel_flag = cancel_flag
        self.sample_rate = sample_rate

    def load_voice_model(self, voice_id, voice_addon_path):
        if voice_id in self.voices:
            return
        tts_onnx = ort.InferenceSession(os.path.join(voice_addon_path, 'data', 'tts.onnx'))
        vocoder_onnx = ort.InferenceSession(os.path.join(voice_addon_path, 'data', 'vocoder.onnx'))
        self.voices[voice_id] = {
            'voice_id': voice_id,
            'voice_addon_path': voice_addon_path,
            'tts_onnx': tts_onnx,
            'vocoder_onnx': vocoder_onnx
        }

    def synthesize_multilingual(self, alternative_synthesizer,
                                text, tts_normalizer_options, chunk_length, overlap_length, log_func=None):
        if alternative_synthesizer._synthesizer is None:
            self.synthesize(text, tts_normalizer_options, chunk_length, overlap_length, log_func=log_func)
        else:
            for segment_lang, segment_text in segment_text_by_language(text):
                if segment_lang == 'mn':
                    self.synthesize(segment_text, tts_normalizer_options, chunk_length, overlap_length, log_func=log_func)
                else:
                    alternative_synthesizer.synthesize(segment_lang, segment_text, log_func=log_func)

    def synthesize(self, text, tts_normalizer_options, chunk_length, overlap_length, log_func=None):
        if self.current_voice is None:
            return

        sonic_processor = sonic.Sonic(self.sample_rate, 1)
        sonic_processor.set_quality(1)
        pitch = self.pitch + self.pitch * self.pitch_offset
        sonic_processor.set_pitch(pitch)
        sonic_processor.set_speed(self.speed)
        sonic_processor.set_volume(self.volume)

        overlap_audio_length = overlap_length * 256
        bartlett = np.bartlett(overlap_audio_length * 2)

        current_voice_data = self.voices[self.current_voice]
        if synthesize_if_possible(self.player, sonic_processor, current_voice_data, text, log_func=log_func):
            return

        tts_onnx, vocoder_onnx = current_voice_data['tts_onnx'], current_voice_data['vocoder_onnx']
        normalizer_time = time.time()
        normalized_texts = normalize_tts_text(dict({"text": text}, **tts_normalizer_options))
        normalizer_time = time.time() - normalizer_time
        if log_func is not None:
            log_func("normalizer time=%.3f text=%s" % (normalizer_time, text))
        for normalized_text in normalized_texts:
            normalized_text = normalized_text.replace("'", "").replace('"', '').strip()
            if normalized_text[-1] not in ".!?":
                normalized_text = normalized_text + '.'
            if len(normalized_text.strip()) == 0:
                continue

            if self.cancel_flag.is_set():
                return

            seq = text_to_sequence(normalized_text)
            text_lengths = np.array([len(seq)], dtype=np.int64)
            seq = np.array([seq], dtype=np.int64)
            mel_time = time.time()
            mel = run_onnx(tts_onnx, [seq, text_lengths, np.array(1.0, dtype=np.float32)])
            mel_time = time.time() - mel_time
            if log_func is not None:
                log_func("mel time=%.3f text=%s" % (mel_time, normalized_text))

            if self.cancel_flag.is_set():
                return

            last_mel_len = mel.shape[2] % chunk_length
            chunk_mels = []
            if last_mel_len < overlap_length * 5:
                slices = [(x + 1) * chunk_length for x in range(mel.shape[2] // chunk_length - 1)]
            else:
                slices = [(x + 1) * chunk_length for x in range(mel.shape[2] // chunk_length)]
            if len(slices) == 0:
                chunk_mels.append(mel)
            else:
                for sl in slices:
                    chunk_mel = mel[:, :, (sl - chunk_length):sl + overlap_length]
                    chunk_mels.append(chunk_mel)
                chunk_mel = mel[:, :, slices[len(slices) - 1]:]
                chunk_mels.append(chunk_mel)

            previous_buffer = []
            for index in range(len(chunk_mels)):
                if self.cancel_flag.is_set():
                    return

                current_chunk_mel = chunk_mels[index]
                vocoder_time = time.time()
                output = run_onnx(vocoder_onnx, [current_chunk_mel])[0, 0, :]
                vocoder_time = time.time() - vocoder_time
                if log_func is not None:
                    log_func("vocoder%i time=%.3f" % (index, vocoder_time))
                audio = (32767 * output).astype(dtype=np.int16)
                buffer_to_save = audio[-overlap_audio_length:]
                if index < len(chunk_mels) - 1:
                    audio = audio[:-overlap_audio_length]
                if len(previous_buffer) > 0:
                    audio[:overlap_audio_length] = audio[:overlap_audio_length] * bartlett[:overlap_audio_length] \
                                                   + previous_buffer * bartlett[overlap_audio_length:]

                previous_buffer = buffer_to_save

                processed_audio = None
                sonic_time = time.time()
                if pitch == 1.0 and self.speed == 1.0 and self.volume == 1.0:
                    processed_audio = audio
                else:
                    chunk_length = 6400

                    def read_processed_samples():
                        processed_chunk_buffer = b""
                        while sonic_processor.samples_available() > 0:
                            processed_chunk_buffer += sonic_processor.read_short(chunk_length)
                        return np.frombuffer(processed_chunk_buffer, dtype=np.int16)

                    n = len(audio) // chunk_length
                    remaining = len(audio) - n * chunk_length
                    for i in range(n):
                        sonic_processor.write_short(audio[i * chunk_length:i * chunk_length + chunk_length].tobytes())
                        processed_chunk_samples = read_processed_samples()
                        processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
                            (processed_audio, processed_chunk_samples))
                    if remaining > 0:
                        sonic_processor.write_short(audio[-remaining:].tobytes())
                        processed_chunk_samples = read_processed_samples()
                        processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
                            (processed_audio, processed_chunk_samples))

                    if index == len(chunk_mels) - 1:
                        sonic_processor.flush()
                        processed_chunk_samples = read_processed_samples()
                        processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
                            (processed_audio, processed_chunk_samples))
                sonic_time = time.time() - sonic_time
                if log_func is not None:
                    log_func("sonic%i time=%.3f" % (index, sonic_time))

                if self.cancel_flag.is_set():
                    return
                self.player.feed(processed_audio.tobytes())
