import os
import wave
import contextlib
import numpy as np

AVAILABLE_WAVS = {
    # alphabets
    'а': 'a.wav', 'А': 'tom_a.wav',
    'б': 'b.wav', 'Б': 'tom_b.wav',
    'в': 'v.wav', 'В': 'tom_v.wav',
    'г': 'g.wav', 'Г': 'tom_g.wav',
    'д': 'd.wav', 'Д': 'tom_d.wav',
    'е': 'ye.wav', 'Е': 'tom_ye.wav',
    'ё': 'yo.wav', 'Ё': 'tom_yo.wav',
    'ж': 'j.wav', 'Ж': 'tom_j.wav',
    'з': 'z.wav', 'З': 'tom_z.wav',
    'и': 'i.wav', 'И': 'tom_i.wav',
    'й': 'hagas_i.wav', 'Й': 'tom_hagas_i.wav',
    'к': 'k.wav', 'К': 'tom_k.wav',
    'л': 'l.wav', 'Л': 'tom_l.wav',
    'м': 'm.wav', 'М': 'tom_m.wav',
    'н': 'n.wav', 'Н': 'tom_n.wav',
    'о': 'o.wav', 'О': 'tom_o.wav',
    'ө': 'oe.wav', 'Ө': 'tom_oe.wav',
    'п': 'p.wav', 'П': 'tom_p.wav',
    'р': 'r.wav', 'Р': 'tom_r.wav',
    'с': 's.wav', 'С': 'tom_s.wav',
    'т': 't.wav', 'Т': 'tom_t.wav',
    'у': 'u.wav', 'У': 'tom_u.wav',
    'ү': 'ou.wav', 'Ү': 'tom_ou.wav',
    'ф': 'f.wav', 'Ф': 'tom_f.wav',
    'х': 'h.wav', 'Х': 'tom_h.wav',
    'ц': 'ts.wav', 'Ц': 'tom_ts.wav',
    'ч': 'ch.wav', 'Ч': 'tom_ch.wav',
    'ш': 'sh.wav', 'Ш': 'tom_sh.wav',
    'щ': 'sch.wav', 'Щ': 'tom_sch.wav',
    'ъ': 'hatuugiin_temdeg.wav', 'Ъ': 'tom_hatuugiin_temdeg.wav',
    'ь': 'zoolnii_temdeg.wav', 'Ь': 'tom_zoolnii_temdeg.wav',
    'ы': 'er_ugiin_i.wav', 'Ы': 'tom_er_ugiin_i.wav',
    # 'ий': 'urt_i.wav', 'Ий': 'tom_urt_i.wav',
    'э': 'e.wav', 'Э': 'tom_e.wav',
    'ю': 'yu.wav', 'Ю': 'tom_yu.wav',
    'я': 'ya.wav', 'Я': 'tom_ya.wav',
    'том': 'tom.wav',

    # digits
    '0': 'teg.wav',
    '1': 'neg.wav',
    '2': 'hoyor.wav',
    '3': 'gurav.wav',
    '4': 'dorov.wav',
    '5': 'tav.wav',
    '6': 'zurgaa.wav',
    '7': 'doloo.wav',
    '8': 'naim.wav',
    '9': 'yos.wav',
    '!': 'anhaarliin_temdeg.wav',
    '"': 'hos_dusal_haalt.wav',
    '#': 'chagt.wav',
    '₮': 'togrog.wav',
    '$': 'dollar.wav',
    '£': 'funt.wav',
    '€': 'euro.wav',
    '¢': 'cent.wav',
    '¥': 'yen.wav',
    '%': 'huvi.wav',
    '&': 'and.wav',
    '\'': 'dusal_haalt.wav',
    '(': 'neeh_haalt.wav',
    ')': 'haah_haalt.wav',
    '+': 'nemeh.wav',
    '-': 'hasah.wav',
    ',': 'taslal.wav',
    '.': 'tseg.wav',
    '/': 'naluu_zuraas.wav',  # '/': 'tashuu_zuraas.wav',
    ':': 'davhar_tseg.wav',
    ';': 'tsegtei_taslal.wav',
    '<': 'bagiin_temdeg.wav',  # '<': 'baga.wav',
    '>': 'ihiin_temdeg.wav',  # '>': 'ih.wav',
    '=': 'tentsuugiin_temdeg.wav',  # '=': 'tentsuu.wav',
    '@': 'et.wav',
    ']': 'haah_dorvoljin_haalt.wav',
    '[': 'neeh_dorvoljin_haalt.wav',
    '\\': 'bohrog_zuraas.wav',
    '^': 'malgai.wav',
    '_': 'dooguur_zuraas.wav',
    '`': 'gedreg_dusal_haalt.wav',
    '{': 'zuun_goy_haalt.wav',
    '}': 'baruun_goy_haalt.wav',
    '~': 'dolgio.wav',
    '№': 'dugaar.wav',  # '№': 'nomer.wav',
    '|': 'bosoo_zuraas.wav',  # '|': 'vertical_chart.wav',
    '?': 'asuultiin_temdeg.wav',
    '*': 'od.wav',
}


def read_wave(path):
    """Reads a .wav file.
    Takes the path, and returns (PCM audio data, sample rate).
    """
    with contextlib.closing(wave.open(path, 'rb')) as wf:
        num_channels = wf.getnchannels()
        assert num_channels == 1
        sample_width = wf.getsampwidth()
        assert sample_width == 2
        sample_rate = wf.getframerate()
        assert sample_rate == 22050
        pcm_data = wf.readframes(wf.getnframes())
        return pcm_data


def synthesize_if_possible(player, sonic_processor, voice_data, text, log_func=None):
    text = text.strip().lower()
    if text not in AVAILABLE_WAVS:
        return False
    wav_file = AVAILABLE_WAVS[text]
    wav_file = os.path.join(voice_data['voice_addon_path'], 'data', 'wavs', wav_file)
    try:
        audio = np.frombuffer(read_wave(wav_file), dtype=np.int16)
    except:
        if log_func is not None:
            log_func("%s doesn't exist in the voice wav data!" % text)
        return False

    processed_audio = None
    chunk_length = 6400

    def read_processed_samples():
        processed_chunk_buffer = b""
        while sonic_processor.samples_available() > 0:
            processed_chunk_buffer += sonic_processor.read_short(chunk_length)
        return np.frombuffer(processed_chunk_buffer, dtype=np.int16)

    n = len(audio) // chunk_length
    remaining = len(audio) - n * chunk_length
    for i in range(n):
        sonic_processor.write_short(audio[i * chunk_length:i * chunk_length + chunk_length].tobytes())
        processed_chunk_samples = read_processed_samples()
        processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
            (processed_audio, processed_chunk_samples))
    if remaining > 0:
        sonic_processor.write_short(audio[-remaining:].tobytes())
        processed_chunk_samples = read_processed_samples()
        processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
            (processed_audio, processed_chunk_samples))

    sonic_processor.flush()
    processed_chunk_samples = read_processed_samples()
    processed_audio = processed_chunk_samples if processed_audio is None else np.concatenate(
        (processed_audio, processed_chunk_samples))

    player.feed(processed_audio.tobytes())
    return True
